use rand::prelude::*;

use std::{
    cell::{Ref, RefCell},
    collections::{HashMap, HashSet},
    rc::{Weak, Rc},
};

use crate::prelude::*;

use fleck_core::crypto::{Keypair, PublicKey};

pub struct InnerRef<'l, T> {
    r: Ref<'l, Option<T>>,
}

impl<'l, T> InnerRef<'l, T> {
    fn build(from: Ref<'l, Option<T>>) -> Option<Self> {
        if from.is_none() {
            None
        } else {
            Some(Self { r: from })
        }
    }
}

impl<'l, T> AsRef<T> for InnerRef<'l, T> {
    fn as_ref(&self) -> &T {
        self.r.as_ref().unwrap()
    }
}

impl<'l, T> std::ops::Deref for InnerRef<'l, T> {
    type Target = T;
    fn deref(&self) -> &Self::Target {
        self.r.as_ref().unwrap()
    }
}

#[derive(Debug,Default)]
pub struct Node {
    pubkey: PublicKey,
    peer: Option<fleck_core::peer::Peer>,
    keypair: RefCell<Option<Keypair>>,
}

impl Node {
    pub fn build_with_peer(peer: fleck_core::peer::Peer) -> Self {
        Self {
            peer: Some(peer),
            ..Default::default()
        }
    }

    pub fn peer(&self) -> Option<fleck_core::peer::Peer> {
        self.peer.clone()
    }

    pub fn pubkey(&self) -> &PublicKey {
        &self.pubkey
    }
    pub(crate) fn keypair(&self) -> Option<InnerRef<'_, Keypair>> {
        InnerRef::build(self.keypair.borrow())
    }

    /*pub fn set_pubkey(&self, pk: PublicKey) {
        let mut pkey = self.pubkey.borrow_mut();
        if pkey.is_some() {
            panic!("Tried to set public key on node that already has a public key!")
        }
        *pkey = Some(pk);
    }

    pub fn set_keypair(&self, kp: Keypair) {
        let mut keyp = self.keypair.borrow_mut();
        if keyp.is_some() {
            panic!("Tried to set keypair on node that already has a keypair!")
        }
        *keyp = Some(kp);
    }

    pub fn gen_keypair(&self) {
        let mut pkey = self.pubkey.borrow_mut();
        let mut keyp = self.keypair.borrow_mut();
        if pkey.is_some() {
            panic!("Tried to generate keypair for node that already has a pubkey!")
        }
        if keyp.is_some() {
            panic!("Tried to generate keypair for node that already has a keypair!")
        }
        let mut csprng = StdRng::from_entropy();
        let kp = Keypair::generate(&mut csprng);
        *pkey = Some(kp.public);
        *keyp = Some(kp);
    }*/
}

pub struct RegTag {}
pub type NodeRegistrationChannel = (RegTag, Rc<Node>);

pub struct NodeService {
    api: Rc<crate::API>,
    all_nodes: RefCell<Vec<Rc<Node>>>,
    // node_by_pubkey: RefCell<HashMap<fleck_core::crypto::PublicKey, Rc<Node>>>,
    node_by_peer: RefCell<HashMap<fleck_core::Peer, Rc<Node>>>,
    self_node: RefCell<Weak<Node>>,
}

impl NodeService {
    pub fn self_node(&self) -> Rc<Node> {
        self.self_node.borrow().upgrade().expect("asking for self node before it's been initialized!")
    }

    pub fn use_ephemeral(&self) {
        
    }

    pub fn node_by_pubkey(&self, pubkey: &PublicKey) -> Option<Rc<Node>> {
        self.all_nodes
            .borrow()
            .iter()
            .find(|n| n.pubkey() == pubkey)
            .map(|r| r.to_owned())
    }

    // nodes we believe we can communicate with directly
    pub fn direct_neighbours(&self) -> Vec<Rc<Node>> {
        self.all_nodes
            .borrow()
            .iter()
            .filter(|n| n.peer.is_some())
            .map(|r| r.to_owned())
            .collect()
    }

    // inform of new nodes 
    /*
    pub fn inform_of_peer(&self, peer: fleck_core::peer::Peer) -> Rc<Node> {
        let mut map = self.node_by_peer.borrow_mut();
        if !map.contains_key(&peer) {
            let node = Rc::new(Node::build_with_peer(peer.clone()));
            map.insert(peer, node.clone());
            self.all_nodes.borrow_mut().push(node.clone());

            self.api.queue::<NodeRegistrationChannel>(node);
        }
        
        map.get(&peer).unwrap().clone()
    }
    */

    /*pub fn inform_of_peer_pubkey(&self, peer: fleck_core::peer::Peer, pkey: fleck_core::crypto::PublicKey) -> Rc<Node> {
        log::trace!("informed of peer {:?} / {:?}", pkey, peer);
        let ret = self.inform_of_peer(peer);
        match ret.pubkey() {
            Some(existing_key) => {
                if *existing_key != pkey {
                    log::warn!("Public key of node has changed!");
                }
            },
            None => ret.set_pubkey(pkey),
        }
        ret
    }*/
}

impl Service for NodeService {
    fn new(api: Rc<crate::API>) -> Self {
        let ret = Self {
            api,
            all_nodes: Default::default(),
            // node_by_pubkey: Default::default(),
            node_by_peer: Default::default(),
            self_node: Default::default(),
        };

        ret
    }

    fn setup(self: &Rc<Self>) {
        self.api.create_channel::<NodeRegistrationChannel>();

        self.api.channel::<fleck_core::ReceivePacketChannel>()
            .sub_ref(fleck_core::ReceiveOrder::Annotate, self, Self::annotate);
    }
}

impl NodeService {
    fn annotate(&self, msg: &mut fleck_core::Message) {
        // annotate node based on peer
        if let Some(peer) = &msg.peer {
            let map = self.node_by_peer.borrow();
            match map.get(&peer) {
                Some(node) => {
                    msg.node = Some(node.clone());
                }
                None => {
                    // the interesting case, it's a new node!
                    // TODO: send them a request for their public key
                },
            }
        }
    }
}
